#include <stdarg.h>
#include "type.h"
#include "inst.h"
#include "inst_con.h"

#include "jni.h"

#include "memory.h"
#include "loader.h"

JavaVMInitArgs	vmargs;
JNIEnv*		env;
JavaVM*		vm;

int initializeExecutionEngine(void)
{
	// JNI

	vmargs.version = 0x00010001;
	JNI_GetDefaultJavaVMInitArgs( &vmargs );
	JNI_CreateJavaVM( &vm, &env, &vmargs );

	// NATIVE

	initNative();

	
	
	return (1);
}

#define CREATE_NATIVE_FRAME(frame,size)					\
	if((frame=allocateMemory(sizeof(type_frame)+sizeof(u4)*size))==0)	\
		return (0);						\
	frame->stack = &frame[1];					\
	frame->tos = frame->stack - 1;					\
	frame->prv = getThreadFrame()

#define CREATE_JAVA_FRAME(frame,size,mt,prv_frame)			\
	if((frame=allocateMemory(sizeof(type_frame)+sizeof(u4)*size))==0)	\
		return (0);						\
	frame->method = mt;						\
	frame->pool = ((type_class*)mt->cls)->pool;			\
	frame->base = ((type_method*)mt->code)->code;			\
	frame->pc = frame->base;					\
	frame->stack = &frame[1];					\
	frame->local = frame->stack+((type_method*)mt->code)->stack;	\
	frame->tos = frame->stack - 1;					\
	frame->prv = prv_frame

	
	
int invokeMainMethod( char* chars, int argc, char* args[] )
{
	u1* name;
	u1* descriptor;
	type_class* class;
	type_method_table* method;
	type_loader* loader;
	type_array* array;
	type_frame* frame;
	int i;
	u1* param;
	type_object* string;
	u4* stack;
	int max;



	printf("invoke main method %s %s %d\n", chars, args[0], argc );
	loader = getBootLoader();

	// class

	name = loadUtf8( chars );

	if( ( class = cload( name, loader ) ) == 0 )
		return (0);

	// 6th ACTIVE USE
	initClass( class );

	name = loadUtf8( "main" );
	descriptor = loadUtf8( "([Ljava/lang/String;)V" );
	
	if( ( method = findMethodName( class, name, descriptor ) ) == 0 )
		return (0);

	// parameter

	name = loadUtf8( "java/lang/String" );
	class = cload( name, loader );
	array = newObjectArray( class, argc );

	for( i = 0 ; i < argc ; i++ )
	{
		((u4*)array->ref)[i] = newString( loadUtf8( args[i] ) );
	}

	CREATE_NATIVE_FRAME( frame, 1 );
	frame->tos++;
	*(frame->tos) = array;

	invokeMethod( method, frame );
	
	return (1);
}

int invokeExternalValist
( type_method_table* method_table, type_object* object, va_list args )
{
	type_frame* frame;
	unsigned int size;
	int i;


	size = ( method_table->parsize > 2 ) ? method_table->parsize : 2 ;
	CREATE_NATIVE_FRAME( frame, size );

	if( method_table->flag & ACC_STATIC )
	{
		printf(" invoke external static\n");
		exit(0);
	}

	frame->tos++;
	*(frame->tos) = object;
	for( i = 1 ; i < method_table->parsize ; i++ )
	{
		frame->tos++;
		*(frame->tos) = va_arg( args, u4 );
	}

	invokeMethod( method_table, frame );

	return (1);
}

int invokeExternal( type_method_table* method_table, ... )
{
	type_frame* frame;
	unsigned int size;
	va_list args;
	int i;


	size = ( method_table->parsize > 2 ) ? method_table->parsize : 2 ;
	CREATE_NATIVE_FRAME( frame, size );

	va_start( args, method_table );
	for( i = 0 ; i < method_table->parsize ; i++ )
	{
		frame->tos++;
		*(frame->tos) = va_arg( args, u4 );
	}
	va_end( args );

	invokeMethod( method_table, frame );

	return (1);
}


int invokeMethod( type_method_table* method_table, type_frame* prv_frame )
{
	type_frame* frame;
	type_method* method;
	type_class* class;
	int size;

	char* native_func_name;
	void* native_func;
	int* native_parameters;
	int native_psize;
	jvalue native_retvalue;


	class = method_table->cls;
	
	/* debug 
	if( strcmp( class->name, "java/lang/ClassLoader" ) == 0 )
	{
		printf("ooc, Class Call \n");
		exit(0);
	}*/
	
	// 2nd ACTIVE USE
	if( class->acc_flag & ACC_STATIC && class->state != CLASSSTATE_COMPLETED )
		initClass( class, 0 );
	
	/*debug*/
	printf("-INVOKE %s %s %s [%d:%c]\n",
		((type_class*)method_table->cls)->name,
		method_table->name, method_table->descriptor,
		method_table->parsize, method_table->retype );
	/*debug*/
	
	method = method_table->code;

	if( method == 0 && ( method_table->flag & ACC_NATIVE ) )
	{
		// native execute

		frame = prv_frame;

		native_func_name = convertNativeFuncName(
			((type_class*)method_table->cls)->name,
			method_table->name, method_table->descriptor );
		
		printf("-NATIVE %s\n", native_func_name );

		if( ( native_func = (void*)findNative(native_func_name) ) == 0 )
		{
			printf("can't find native functions\n");
			exit(0);
		}

		native_psize = method_table->parsize;
		
		if( method_table->flag & ACC_STATIC )
		{
			native_psize+=2;
			native_parameters = malloc( sizeof(int)*native_psize );
			native_parameters[0] = env;
			native_parameters[1] = method_table->cls;
			frame->tos -= (native_psize-2);
			memcpy( &native_parameters[2], frame->tos+1,
				sizeof(int)*(native_psize-2) );
		}
		else
		{
			native_psize++;
			native_parameters = malloc( sizeof(int)*native_psize );
			native_parameters[0] = env;
			frame->tos -= (native_psize-1);
			memcpy( &native_parameters[1], frame->tos+1,
				sizeof(int)*(native_psize-1) );
		}

		invokeNativeMethod( native_func, native_psize, native_parameters,
			method_table->retype, &native_retvalue );

		switch( method_table->retype )
		{
		case 'V':
			break;
		case 'D':
			PUSH_DOUBLESIZE( (u4*)&(native_retvalue.d) );
			break;
		case 'J':
			PUSH_DOUBLESIZE( (u4*)&(native_retvalue.j) );
			break;
		case 'F':
			push( native_retvalue.f );
		default:
			push( native_retvalue.i );
			break;
		}
		
		printf("-NATIVE END %s\n", native_func_name );

	}
	else
	{
		// java execute

		size = method->stack + method->local;
		CREATE_JAVA_FRAME( frame, size, method_table, prv_frame );
		/* debug */
		frame->lsize = method->local;
		frame->ssize = method->stack;
		/* debug */
		updateThreadFrame( frame );
		
		if( method_table->parsize > 0 )
		{
			prv_frame->tos -= method_table->parsize;
			memcpy( frame->local, prv_frame->tos+1, sizeof(u4)*method_table->parsize );
		}
	
		interpret( frame );

		updateThreadFrame( prv_frame );
	}

	/*debug */
	printf("-END %s %s %s [%d:%c]\n",
		((type_class*)method_table->cls)->name,
		method_table->name, method_table->descriptor,
		method_table->parsize, method_table->retype );
	/*debug*/
	
	return (1);
}




int interpret( type_frame* frame )
{
	u1 inst;
	u2 exception_jump;
	int i;
	u4* j;

	while(1)
	{
		inst = *(frame->pc);

		
		/* debug */
		printf("  [interpret] %x ", inst );
		for( i = 0 ; i < frame->lsize ; i++ )
			printf( "[%x]", (frame->local)[i] );
		for( j = frame->stack ; j <= frame->tos ; j++ )
			printf( "{%x}", *j );
		printf("\n");
		/* debug */

		
		inst_map[inst](frame);
		frame->pc += insnLen[inst];

		if( insnLen[inst] == 0 )
//			if( frame->prv != 0 )
//				((type_frame*)frame->prv)->exception = frame->exception;
			return (1);

		if( frame->exception != 0 )
		{
			exception_jump = findException( frame );
			
			if( exception_jump == 0 )
			{
				if( frame->prv == 0 )
					exit(0);

				((type_frame*)frame->prv)->exception = frame->exception;
				((type_frame*)frame->prv)->tos++;
				*((type_frame*)frame->prv)->tos = frame->exception;

//				printf("THROW FRAME POP\n");
				return (1);
			}
			else
			{
				frame->exception = 0;
				frame->pc = frame->base + exception_jump;
			}
		}		
	}
	
//	printf( "interpret\n" );
	return (0);
}
