#include "hash.h"

#define INIT_SIZE		0x4000
#define NEED_RESIZE(tab)	( 4 * (tab)->count >= 3 * (tab)->size )
#define LIST_STEP(hash)		( 8 * (hash) + 7 )

static int	hashFindSlot(hash_t, const void*);
static hash_t	hashResize(hash_t);

struct _hashtab {
	const void	**list;
	int		count;
	int		size;

	pf_comp		comp;
	pf_hash		hash;
	pf_alloc	alloc;
	pf_free		free;
};

hash_t hashInit(pf_hash hash, pf_comp comp, pf_alloc alloc, pf_free free)
{
	hash_t tab;

	if( alloc == 0 )
		tab = (hash_t)malloc( sizeof( *tab ) );
	else
		tab = alloc(sizeof(*tab));

	if( tab == 0 ) return (0);

	tab->hash = hash;
	tab->comp = comp;
	tab->alloc = alloc;
	tab->free = free;

	hashResize(tab);
	return (tab);
}

void hashFinal(hash_t tab)
{
	int i;
	for( i = 0 ; i < tab->size ; i++ )
	{
		if( tab->list[i] != 0 )
			hashRemove( tab, tab->list[i] );
	}

	if( tab->free )
	{
		tab->free(tab->list);
		tab->free(tab);
	}
	else
	{
		free(tab->list);
		free(tab);
	}
}

void* hashAdd(hash_t tab, const void* ptr)
{
	int 	index;
	void	*rtn;


	if( NEED_RESIZE(tab) )
	{
		printf("HASH memory shortage %x %x \n", tab->size, tab->count );
		exit(0);
		
		if( hashResize(tab) == 0 )
			return (0);
	}

	index = hashFindSlot( tab, ptr );

	if( tab->list[index] == 0 )
	{
		tab->list[index] = ptr;
		tab->count++;
	}
	else
	{
		printf("ߺ\n");
		exit(0);
	}
	
	rtn = (void*)tab->list[index];
	return (rtn);
}

void hashRemove(hash_t tab, const void* ptr)
{
	int	index;
	
	index = hashFindSlot( tab, ptr );
	if( tab->list[index] != 0 && tab->list[index] == ptr )
	{
		tab->count--;
		tab->list[index] = 0;
	}
}

void* hashFind(hash_t tab, const void* ptr)
{
	int	index;
	void*	rtn;

//	printf("enter the find hash\n");
	
	index = hashFindSlot( tab, ptr );
	rtn = (tab->list[index] == 0 ) ? 0 : (void*)tab->list[index];
	return (rtn);
}

static int hashFindSlot(hash_t tab, const void* ptr)
{
	const int hash = (*tab->hash)(ptr);
	const int startIndex = hash & (tab->size - 1);
	const int step = LIST_STEP(hash);

	int index;
	int deletedIndex = -1;

//	printf("[index] %x (size:%x/step:%x)\n", startIndex, tab->size, step );

	if( ptr == 0 )
		return 0;

	index = startIndex;
	for(;;)
	{
		const void **const ptr2 = &tab->list[index];

		if( *ptr2 == 0 )
		{
//			printf("[exit] PTR2 NULL ");
			return (deletedIndex >= 0) ? deletedIndex : index ;
		}

		else if( *ptr2 == ptr || (*tab->comp)(ptr,*ptr2) == 0 )
		{
//			printf("[exit] FIND ");
			return (index);
		}

		index = (index + step) & (tab->size - 1);

		if( index == startIndex )
		{
			if( deletedIndex >= 0 )
				return( deletedIndex );
			exit(0);
		}
	}

//	printf("----- --- find slot not found\n");
}

static hash_t hashResize(hash_t tab)
{
	const int size = (tab->size > 0) ? (tab->size * 2) : INIT_SIZE;
	const void **list;
	int index;


	if( tab->alloc )
		list = tab->alloc( size * sizeof(*list) );
	else
		list = malloc( size * sizeof(*list) );

	if( !NEED_RESIZE(tab) )
	{
		if( tab->free )	tab->free( list );
		else free( list );
		return (tab);
	}

	if( list == 0 )	return (0);

	if( tab->free )	tab->free( tab->list );
	else free( tab->list );
	
	tab->list = list;
	tab->size = size;

	return (tab);
}
