//
//	jar file operation
//

#include <stdio.h>
#include <string.h>
#include "jar.h"

#include "inflate.h"
//#include "../typendata/macros.h"
//#include "../gc/gc.h"
//#include "../log/log.h"

// local definations
#define EF_sig		('P' + ('K'<<8) + (5<<16) + (6<<24)) // end field signature
#define CE_sig		('P' + ('K'<<8) + (1<<16) + (2<<24)) // central dir signature
#define LO_sig		('P' + ('K'<<8) + (3<<16) + (4<<24)) // local header signature
#define SOL			30	// size of local header field
#define STORED		0	// definations of
#define DEFLATED	8	// compression method

// local function's prototypes
int read_end_field(jFILE* fd);
int link_from_loc_field(jFILE* fd);

// <global>
// get class data from jar file
// paramater	: jFILE pointer, jar* entry
// return		: 0 (false)
//				  u1* (greped data from jar file)
//
u1*
get_jar_data(jFILE* fd, jar* entry)
{
	u1* data;
	u1* buf;
	class_file* src = fd->fp;

	switch(entry->comp_meth) {
	case STORED:
		if (entry->comp_size != entry->ucomp_size)
		{
//			Log("get_jar_data", "wrong compressed size or decompressed size");
			return 0;
		}
		data = malloc(entry->ucomp_size);
		fseek(src->fd,entry->cur,SEEK_SET);
		readm(data,entry->ucomp_size,sizeof(u1),src);
		break;
	case DEFLATED:
		// not yet
		buf = malloc(entry->comp_size);
		fseek(src->fd,entry->cur,SEEK_SET);
		readm(buf,entry->comp_size,sizeof(u1),src);

		data = malloc(entry->ucomp_size);
		if (data && inflate_oneshot(buf, entry->comp_size, data, entry->ucomp_size)==0)
		{
			// done
			free(buf);
			return data;
		}
		else
		{
			// error
			free(buf);
			free(data);
			return 0;
		}
		break;
	default:
//		Log("get_jar_data", "can not identify compression method.");
		return 0;
	}
	return data;
}

// <global>
// search class from jar list
// paramater	: jFILE pointer, const char* class name
// return		: 0 (false)
//				  jar* (found class and return jar entry node)
//
jar*
search_jar(jFILE* fd, const char* name)
{
	class_file* file = fd->fp;
	char* tmp = malloc(strlen(name)+7);
	int cnt = fd->cnt;
	jar* entry = fd->head;

	memcpy(tmp,name,strlen(name));
	memcpy(tmp+strlen(name),".class",6);
	tmp[strlen(name)+6] = '\0';
//	tmp[strlen(name)] = '\0';

	if((fd->head==0) || (fd->cnt==0))
		return 0;

	while(cnt>0)
	{
		if(strcmp(entry->name,tmp)==0)
		{
			free(tmp);
			return entry;
		}

		entry = entry->next;
		cnt--;
	}

	free(tmp); // free temporary name string

	return 0;
}

// <global>
// open jar file and link class info list.
// paramater	: const char* name (jar file name)
// return		: 0 (false: file not found)
//				  jFILE pointer
//
jFILE*
open_jar(char* name)
{
	jFILE* fd = malloc(sizeof(jFILE));

	long cnt=0;
	FILE *fp;
	class_file* hd = malloc(sizeof(class_file));
	hd->type = 0;

	fp = fopen(name,"r+b");

	if (fp==0)
	{
		free(fd,0);
		free(hd,0);
		return 0;
	}

	fseek(fp,0,SEEK_END);
	cnt = ftell(fp);
	fseek(fp,0,SEEK_SET);

	hd->fd = fp;
	hd->size = cnt;
	fd->fp = hd;

//	read end field
	if (read_end_field(fd)==0)
	{
		free(fd,0);
		free(hd,0);
		return 0;
	}
//Log("open_jar","before link from loc field");
//	link class info from central field
	link_from_loc_field(fd);
//Log("open_jar","after ling from loc field");

	return (fd);
}

// <local>
// link class info from local field
// paramater	: jFILE pointer
// return		: 0 (false)
//				  1 (true)
//
int
link_from_loc_field(jFILE* fd)
{
	class_file* file = fd->fp;
	u2 name_len, ex_len, method;
	u4 com_size, ucom_size;
	u4 base = fd->loc_entry;
	u4 EOL = fd->cen_entry;
	u4 pos = base;
	u4 sig, crc;
	char* tmp;
	jar* node=0;
	jar* cur;
	int cnt=0; // enrty count

	while(pos<EOL)
	{
		fseek(file->fd, pos, SEEK_SET); // set entry position.

		// verify central dir signature.
		rreadu4(sig, file);
		if(sig != LO_sig)
			return 0; // TODO : free all nodes

		seekm(file, 4);
		rreadu2(method, file);
		seekm(file, 4);
		rreadu4(crc, file);

		rreadu4(com_size, file);
		rreadu4(ucom_size, file);
		rreadu2(name_len, file);
		rreadu2(ex_len, file);

//		Log("link from central field","name length = %d",name_len);
//		Log("link from central field","ex length = %d",ex_len);

//		seekm(file, 12); // skip

//////////////////////////////////////////////////////////////////
		cur = malloc(sizeof(jar)); // create new jar node.
		cur->next = 0;
		if(node==0)
			fd->head = cur;
		else
			node->next = cur;

		node = cur;
//////////////////////////////////////////////////////////////////

		if (name_len>0) {
		tmp = malloc(name_len+1);
		readm(tmp, name_len, sizeof(char), file); // read name
		tmp[name_len] = '\0';
		}

		cur->name = tmp;

		cur->comp_meth = method;
		cur->comp_size = com_size;
		cur->ucomp_size = ucom_size;
		cur->cur = pos + SOL + ex_len + name_len;
		cur->crc = crc;
//		Log("link from central field","linking %s",tmp);

		// go to next entry
		pos += SOL + ex_len + name_len + com_size;
		cnt++;
	}

	fd->cnt = cnt;
//	Log("link from central field", "linked %d class(es)", cnt);
	return 1;
}

// <local>
// read end field info of jar file
// paramater	: jFILE pointer
// return		: 0 (false)
//				  1 (true)
//
int
read_end_field(jFILE* fd)
{
	class_file* file = fd->fp;
	long EF_size = 22;
	u4 current_pos, cen_size, loc_size, sig=0;

//	Log("jar.c","EF_sig [ %d ]",EF_sig);
//	Log("jar.c","LO_sig [ %d ]",LO_sig);

// find end field signature
	fseek(file->fd, file->size - EF_size, SEEK_SET);
	rreadu4(sig, file);

	switch((char)sig) { // fix me : calling seekm by maximum 3 times is too wasty.
	case 6 :
		seekm(file,-1); // go back 3 bytes
	case 5 :
		seekm(file,-1); // go back 2 bytes
	case 'K' :
		seekm(file,-1); // go back a byte
		seekm(file,-4); // rewind u4

		rreadu4(sig,file); // reload

	case 'P' :
		// verify local header signature
		if (sig != EF_sig)
		{
			return 0;
		}

		// found end field entry
		current_pos = ftell(file->fd);

		seekm(file,8);

		rreadu4(cen_size,file);
		rreadu4(loc_size,file);

//		Log("jar.c","cen_size [ %d ]",cen_size);
//		Log("jar.c","loc_size [ %d ]",loc_size);

		fd->cen_entry = current_pos - cen_size - 4;
		fd->loc_entry = fd->cen_entry - loc_size;

		// verify local entry point
		fseek(file->fd, fd->loc_entry, SEEK_SET);
		rreadu4(sig,file);
		if (sig != LO_sig)
		{
			return 0;
		}

		// verify central dir entry point
		fseek(file->fd, fd->cen_entry, SEEK_SET);
		rreadu4(sig,file);
		if (sig != CE_sig)
		{
			return 0;
		}

		break;
	} // end of switch

	return 1;
}
