#include "type.h"
#include "jar.h"
#include "jtypes.h"

#include "memory.h"
#include "loader.h"


type_loader* boot_loader;
type_class* class_object;
type_class* class_string;
type_class* class_class;

u1* string_constant_value;
u1* string_deprecated;
u1* string_synthetic;
u1* string_code;
u1* string_exceptions;


type_class* findClass( type_object* string )
{
	type_class* result;
	type_array* array = *(u4*)string->ref;
	jchar* text = (jchar*)array->ref;
	char name[256];
	char* n;
	int i;


	for( i = 0 ; i < array->length ; i++ )
	{
		if( ( name[i] = (char)text[i] ) == '.' )
			name[i] = '/';
	}
	name[i] = '\0';

	result = cload( loadUtf8(name), 0 );

	return(result);
}

// ============================================================================
// INITIALIZE CLASS LOADER
// ============================================================================

int initializeClassLoader(void)
{
	type_class* class;
//	field_list* field;
	int size;
	char* str;


	string_constant_value = loadUtf8( STRING_CONSTANT_VALUE );
	string_deprecated = loadUtf8( STRING_DEPRECATED );
	string_synthetic = loadUtf8( STRING_SYNTHETIC );
	string_code = loadUtf8( STRING_CODE );
	string_exceptions = loadUtf8( STRING_EXCEPTIONS );


	str = loadUtf8("java/lang/ClassLoader");
	loadClassLoader( str, 0 );

/*	
	str = interUTF8new("java/lang/Object");
	class_object = (type_class*)cload( str, boot_loader );
	str = interUTF8new("java/lang/Class");
	class_class = (type_class*)cload( str, boot_loader );
	str = interUTF8new("java/lang/String");
	class_string = (type_class*)cload( str, boot_loader );
*/
	return (1);
}


type_loader* getBootLoader(void)
{
	return (boot_loader);
}

// ============================================================================
// loads a class loader.
// p1 : utf8->data
// p2 : class loader
// ============================================================================
type_loader* loadClassLoader(u1* name, type_loader* base_loader)
{
	type_class* loader;
	type_method_table* method;
	class_file* fp;
	u1* str;


	if( ( loader = allocateMemory( sizeof( type_loader ) ) ) == 0 )
		return (0);

	((type_loader*)loader)->path = initClassPath();

	if( base_loader == 0 )
	{	// boot loader
		printf("BOOTLOADER=%x\n",loader);
		base_loader = loader;
		boot_loader = loader;
	}

	loader->cls = cload( loadUtf8("java/lang/ClassLoader"), base_loader );
	// do you want to invoke <init>?
	loader->ref = allocateMemory( ((type_class*)loader->cls)->intsize );

	// execution engine
/*	method = findMethodName( loader->cls,
		interUTF8new("<init>"), interUTF8new("(Ljava/lang/ClassLoader;)V") );
	invokeExternal( method, 0 );
*/		
	return (loader);
}


// ============================================================================
// loads a class.
// p1 : utf8->data
// p2 : class loader
// ============================================================================
type_class* cload(u1* name, type_loader* loader)
{
	type_class* cls;
	class_file* fp;
	u1* str;



	if( loader == 0 )
		loader = boot_loader;
	
	if( ( cls = findFromList( name, loader ) ) != 0 )
		return (cls);

	if( name[0] == '[' ) return (cloadArray( name, loader ));
	if( name[1] == '\0' ) return (cloadPrime( name, loader ));

	if( ( fp = findFromFile( name, loader ) ) == 0 )
		return (0);

	printf("[CL] {%s} loading...\n", name );

	if( ( cls = allocateMemory( sizeof( type_class ) ) ) == 0 )
		return (0);

	cls->state = CLASSSTATE_UNLOADED;
	cls->loader = loader;
	cls->name = name;

	if( cls && fp )
	{
		readClassData( cls, fp );
		linkClassList( cls, loader );
	}

	cls->state = CLASSSTATE_LOADED;
//	printf( "[CL] {%s} loaded\n", name );

	resolveClass( cls, loader );
//	printf( "[CL] {%s} resolved\n", name );

	class_class = cload( loadUtf8("java/lang/Class"), loader );
	cls->cls = class_class;
	// do you want to invoke <init>?
	cls->ref = allocateMemory( class_class->intsize );

	return (cls);
}

type_class* cloadArray(u1* name, type_loader* loader)
{
	type_class* cls;
	u1* str;

	if( ( cls = allocateMemory( sizeof( type_class ) ) ) == 0 )
		return (0);

	cls->loader = loader;
	cls->name = name;
	cls->super = cload( loadUtf8("java/lang/Object"), loader );

	cls->ninterface = 2;
	cls->interfaces = allocateMemory( sizeof(u4*)*2 );
	cls->methods = 0;
	cls->nmethod = 0;
	
	str = loadUtf8( "java/lang/Cloneable" );
	cls->interfaces[0] = cload( str, loader );
	str = loadUtf8( "java/io/Serializable" );
	cls->interfaces[1] = cload( str, loader );

	linkClassList( cls, loader );

	cls->state = CLASSSTATE_COMPLETED;
	cls->cls = cload( loadUtf8("java/lang/Class"), loader );
	
	return (cls);
}

type_class* cloadPrime(u1* name, type_loader* loader)
{
	type_class* cls;


	if( ( cls = allocateMemory( sizeof( type_class ) ) ) == 0 )
		return (0);

	cls->loader = loader;
	cls->name = name;
	cls->super = cload( loadUtf8("java/lang/Object"), loader );

	linkClassList( cls, loader );
	
	cls->state = CLASSSTATE_COMPLETED;
	cls->cls = cload( loadUtf8("java/lang/Class"), loader );

	return (cls);
}


int resolveClass( type_class* class, type_loader* loader)
{

	type_constant* pool;
//	type_class* tmp_class;
	int i;
	u4 idx;
	u4 cidx;
	u4 tidx;
	u4 nidx;
	u4 didx;

	u2* ptr2;
	u1* str;

	type_method_table* tmp_methods;

	pool = class->pool;

	for( i = 1 ; i < pool->size ; i++ )
	{
		
//		printf( "  [POOL] %x %x %x\n", i, (pool->data)[i], (pool->tag)[i] );

		switch( (pool->tag[i]) )
		{
		case 0x17:

			cidx = (pool->data)[i];
			(pool->data)[i] = cload( (pool->data)[cidx], loader );
			(pool->tag)[i] &= CONSTANT_RESOLVED;
			break;

		case 0x18:

			cidx = (pool->data)[i];
			(pool->data)[i] = newString( (pool->data)[cidx] );
			(pool->tag)[i] &= CONSTANT_RESOLVED;
			break;

		case 0x19:

			idx = (pool->data)[i];
			ptr2 = &idx;
			cidx = ptr2[0];
			nidx = (pool->data)[cidx];
			tidx = (pool->data)[ptr2[1]];
			ptr2 = &tidx;

			if( (pool->tag)[cidx] & CONSTANT_UNRESOLVED )
			{
				(pool->data)[cidx]
					= cload( (pool->data)[nidx], loader );
				(pool->tag)[cidx] &= CONSTANT_RESOLVED;
			}
			
			(pool->data)[i] = findFieldName( (pool->data)[cidx],
				(pool->data)[ptr2[0]], (pool->data)[ptr2[1]]);
			(pool->tag)[i] &= CONSTANT_RESOLVED;

			/* debug */
			if( (pool->data)[i] == 0 )
			{
				printf("unkowned field\n");
				exit(0);
			}
			/* debug */

			break;

		case 0x1a:
			
			idx = (pool->data)[i];
			ptr2 = &idx;
			cidx = ptr2[0];
			nidx = (pool->data)[cidx];
			tidx = (pool->data)[ptr2[1]];
			ptr2 = &tidx;

			if( (pool->tag)[cidx] & CONSTANT_UNRESOLVED )
			{
				(pool->data)[cidx]
					= cload( (pool->data)[nidx], loader );
				
				(pool->tag)[cidx] &= CONSTANT_RESOLVED;
			}

			(pool->data)[i] = findMethodName( (pool->data)[cidx],
				(pool->data)[ptr2[0]], (pool->data)[ptr2[1]]);
			(pool->tag)[i] &= CONSTANT_RESOLVED;
			
			/* debug */
			if( (pool->data)[i] == 0)
			{
				printf("unkowned mehotd\n");
				exit(0);
			}
			/* debug */

			break;

		case 0x1b:
			
			idx = (pool->data)[i];
			ptr2 = &idx;
			cidx = ptr2[0];
			nidx = (pool->data)[cidx];
			tidx = (pool->data)[ptr2[1]];
			ptr2 = &tidx;

			if( (pool->tag)[cidx] & CONSTANT_UNRESOLVED )
			{
				(pool->data)[cidx]
					= cload( (pool->data)[nidx], loader );
				
				(pool->tag)[cidx] &= CONSTANT_RESOLVED;
			}

			(pool->data)[i] = findInterfaceMethodName( (pool->data)[cidx],
				(pool->data)[ptr2[0]], (pool->data)[ptr2[1]]);
			(pool->tag)[i] &= CONSTANT_RESOLVED;
			
			/* debug */
			if( (pool->data)[i] == 0)
			{
				printf("unkowned interface mehotd\n");
				exit(0);
			}
			/* debug */
			
		default:
		
			break;	
		}

//		printf( "  [POOL] %x %x %x\n", i, (pool->data)[i], (pool->tag)[i] );
	}

	class->state = CLASSSTATE_RESOLVED;
	return (1);
}


int initClass( type_class* class, type_loader* loader )
{
	type_class* super;
	type_method_table* method;
	u1* name;
	u1* descriptor;


	class->state = CLASSSTATE_COMPLETED;
	
	printf( "<clinit> %s\n", class->name );

//	if( class->state == CLASSSTATE_LOADED );
//		resolveClass( class, loader );
	
	super = class->super;
	if( super != 0 && super->state != CLASSSTATE_COMPLETED )
		initClass( super, loader );

	name = loadUtf8( "<clinit>" );
	descriptor = loadUtf8( "()V" );

	if( ( method = findMethodName( class, name, descriptor ) ) == 0 )
		return (0);

	invokeMethod( method, 0 );

//	class->state = CLASSSTATE_COMPLETED;
	return (1);
}


classpath_list* linkClassPath(classpath_list* prv, char *name, int type)
{
	classpath_list* path;
	int size;


	path = malloc( sizeof(classpath_list) );
	// should be use newUTF()
	size = strlen(name) + 1;
	path->path = malloc(size);
	memcpy(path->path, name, size);
	path->path[size] = '\0';
	path->type = type;
	path->link = 0;
	path->next = 0;

	/* control list
	if( prv == 0 )
		return (path);
	if( prv->next != 0 )
		path->next = prv->next;
	prv->next = path;
	*/
	path->next = prv;
	
	return (path);
}

classpath_list* initClassPath(void)
{
	classpath_list* path;


	path = linkClassPath( NULL, "./", PATH_DIR );
	path = linkClassPath( path, DEF_CLASSLIBRARY, PATH_ZIP );

	return (path);
}



class_file* findFromFile(u1* name, type_loader* loader)
{
	classpath_list* path;
	class_file* hd;
	jFILE* jfp;
	jar* entry;
	char* fullname;
	char* tmp;
	FILE* fp;

      
	for( path = loader->path ; path != 0 ; path = path->next )
	{
		switch( path->type )
		{
		case PATH_ZIP:

			if( path->link == 0 )
			{
				if( ( jfp = open_jar( path->path ) ) == 0 )
					break;
				path->link = jfp;
			}
			else
				jfp = path->link;
			if( ( entry = search_jar( jfp, name ) ) == 0 )
				break;
			hd = malloc( sizeof( class_file ) );
			hd->type = 1;
			if( ( hd->buf = get_jar_data( jfp, entry ) ) == 0 )
				break;
			hd->size = entry->ucomp_size;

			return (hd);

		case PATH_DIR:

			fullname = malloc( strlen(path->path)+strlen(name)+7 );
			tmp = fullname;
			memcpy( tmp, path->path, strlen(path->path) );
			tmp = tmp + strlen( path->path );
			memcpy( tmp, name, strlen(name) );
			tmp = tmp + strlen( name );
			memcpy( tmp, ".class\0", 7 );
			fp = fopen( fullname, "rb" );

			if( fp )
			{
				hd = malloc( sizeof( class_file ) );
				hd->type = 0;
				hd->fd = fp;
				fseek( fp, 0, SEEK_END );
				hd->size = ftell( fp );
				fseek( fp, 0, SEEK_SET );
			}
			else
			{
				hd = 0;
			}

			return (hd);

		default:

			break;
		}
	}

	return (0);
}


void linkClassList(type_class* class, type_loader* loader)
{
	class->next = loader->list;
	loader->list = class;
	return;
}

/*
 * find from list of class laoder
 *
 * name : utf8->data
 * loader : class loader
 */
type_class* findFromList(u1* name, type_loader* loader)
{
	type_class* class;

	
	if( loader == 0 )
	{
		printf("[CL] no class loader\n");
		return (0);
	}

	if ( name == 0 )
	{
		printf("[CL] no class name\n");
		return (0);
	}

	class = loader->list;

	for( ; class != NULL ; class = class->next )
	{
		if( class->name == name )
			return (class);
	}

	return (0);
}



/*=========================================================================
 * FUNCTION:      read_ version_info()
 * TYPE:          private class file load operation
 * OVERVIEW:      Read the first few bytes of a Java class file,
 *                checking the file type and version information.
 * INTERFACE:
 *   parameters:  classfile pointer
 *   returns:     <nothing>
 *=======================================================================*/

void readVersionInfo(class_file* fp)
{
	u2 minor_ver;
	u2 major_ver;
	u4 magic;


	readu4( magic, fp );
	readu2( minor_ver, fp );
	readu2( major_ver, fp );

	return;
}

void readConstantPool(type_class* class, class_file* fp)
{
	u4 cp_cnt;
	type_constant* pool;
	u1* tag;
	u4* data;	
	u1 buf1;
	u2 buf2;
	u4 buf4;
	u1* ptr1;

	char* str;
	utf8* utf;
	u2 idx1;
	u2 idx2;
	int i;


	readu2( cp_cnt, fp );

	if( ( pool = allocateMemory(
		sizeof(type_constant) + sizeof(u4)*(cp_cnt+1) + sizeof(u1)*(cp_cnt+1)
		) ) == 0 )
       	{
		printf("[CL] can't read constant pool\n");
		return;
	}

	class->pool = pool;
	data = pool+1;
	tag = data+(cp_cnt+1);

	pool->size = cp_cnt; 
	pool->data = data;
	pool->tag = tag;
	
	data[0] = 0;
	tag[0] = 0;

	for( i = 1 ; i < cp_cnt ; i++ )
	{
		
		readu1( buf1, fp );
		tag[i] = buf1;
		
		switch( tag[i] )
		{
		case CONSTANT_UTF8:

			readu2( buf2, fp );
			if( buf2 < 0 )
			{
				printf("[CP] string buffer error\n");
				exit(0);
			}

			/* not completed */
			if( ( str = malloc( sizeof(u1)*(buf2+1) ) ) == 0 )
				exit(0);
			if( buf2 > 0 )
				readm( str, buf2, sizeof(u1), fp );
			str[buf2] = '\0';
			str = loadUtf8( str );
			data[i] = (u4)str;

			if( buf2 != strlen(str) )
				printf(" differnce data\n");
	
			/* debug */
//			printf("[CL] POOL-%x\t UTF8 {%s}\n",i,data[i]);

			break;

		case CONSTANT_CLASS:
		case CONSTANT_STRING:

			readu2( buf2, fp );
			data[i] = buf2;
			tag[i] |= CONSTANT_UNRESOLVED;
			
			/* debug */
//			printf("[CL] POOL-%x\t CLASS/STRING {%x}\n",i,data[i]);
			
			break;

		case CONSTANT_METHODREF:
		case CONSTANT_FIELDREF:
		case CONSTANT_INTERFACEMETHODREF:
		case CONSTANT_NAMEANDTYPE:

			readu2( idx1, fp );
			readu2( idx2, fp );
			data[i] = (idx2<<16) | (idx1);
			tag[i] |= CONSTANT_UNRESOLVED;

			/* debug */
//			printf("[CL] POOL-%x\t UTF8-REF {%x}\n",i,data[i]);
			break;
			
		case CONSTANT_INT:
		case CONSTANT_FLOAT:

			readu4( buf4, fp );
			ptr1 = &buf4;
			data[i] = (ptr1[3]<<24)|(ptr1[2]<<16)|(ptr1[1]<<8)|(ptr1[0]);
			/* debug */
//			printf("[CL] POOL-%x\t 1-WORD {%f}\n",i,data[i]);
			break;

		case CONSTANT_LONG:
		case CONSTANT_DOUBLE:

			readu4( buf4, fp );
			ptr1 = &buf4;
			data[i] = (ptr1[3]<<24)|(ptr1[2]<<16)|(ptr1[1]<<8)|(ptr1[0]);
			/* debug */
//			printf("[CL] POOL-%x\t 2-WORD-U {%x}\n",i,data[i]);
			i++;
			readu4( buf4, fp );
			ptr1 = &buf4;
			data[i] = (ptr1[3]<<24)|(ptr1[2]<<16)|(ptr1[1]<<8)|(ptr1[0]);
			/* debug */
//			printf("[CL] POOL-%x\t 2-WORD-D {%x}\n",i,data[i]);
			break;

		default:
			break;
		}
	}

//	verifyConstant(class);
	
	return;
}

void readClassInfo( type_class* class, class_file* cfile )
{
	u2 buf2;
	u4 buf4;
	u1* super;
	u4* data;
	type_constant* pool;
	

	pool = class->pool;

	// class flag	
	readu2( buf2, cfile );
	class->acc_flag = buf2;
	
	// this class (name)
	readu2( buf2, cfile );
	buf4 = (pool->data)[buf2];
	class->name = (pool->data)[buf4];
	(pool->data)[buf2] = class;
	(pool->tag)[buf2] &= CONSTANT_RESOLVED;

	// super class
	readu2( buf2, cfile );
	if( buf2 == 0 )
	{
		class->super = 0;
	       	return;
	}
	else
	{
		buf4 = (pool->data)[buf2];
		super = (pool->data)[buf4];
		class->super = cload( super, class->loader );
		(pool->data)[buf2] = class->super;
		(pool->tag)[buf2] &= CONSTANT_RESOLVED;
	}

	return;
}

void readInterfaces( type_class* class, class_file* cfile )
{
	u2 buf2;
	u4 buf4;
	u1* name;
	type_class* interface;
	u4* table;
	type_constant* pool;
	int i;
	u2 cnt;

	
	readu2( cnt, cfile );
	if( cnt == 0 )
	{
		class->ninterface = 0;
		class->interfaces = 0;
		return;
	}

	class->ninterface = cnt;
	class->interfaces = allocateMemory( sizeof(u4)*cnt );
	table = class->interfaces;

	pool = class->pool;
	for( i = 0 ; i < cnt ; i++ )
	{
		readu2( buf2, cfile );
		buf4 = (pool->data)[buf2];
		name = (pool->data)[buf4];
		interface = cload( name, class->loader );

		// interface table
		table[i] = interface;

		// set in constant pool
		(pool->data)[buf2] = interface;
		(pool->tag)[buf2] &= CONSTANT_RESOLVED;
	}
	
	return;
}



type_method* readAttributeCode( type_class* class, class_file* cfile )
{
	type_method* method;
	u4 length;
	u4 code_length;
	u2 stack;
	u2 local;
	u1* code;

	u1 buf1;
	int i;
	u2* tmp;	

	readu4( length, cfile );

	if( length <= 0 )
		return (0);

	readu2( stack, cfile );
	readu2( local, cfile );
	readu4( code_length, cfile );
	length = length - 8;

	method = allocateMemory( sizeof(type_method) + length );

	method->stack = stack;
	method->local = local;
	code = (u1*)method + sizeof(type_method);
	readm( code, length, sizeof(u1), cfile );
	method->code = code;
	method->exception_length = *(u2*)(code+code_length+1);

	if( method->exception_length == 0 )
		method->exception = 0;
	else
	{
		method->exception = code+code_length+3;
		// convert : big endian >> small endian
//		tmp = method->exception;
//		for( i = (method->exception_length*4) ; i > 0  ; i-- )
//			*tmp = ( (*tmp) & 0xff00 ) | ( (*tmp) & 0x00ff ); 
	}

	/* not completed */

	return (method);
}


type_method_exception* readAttributeExceptions
( type_class* class, class_file* cfile )
{
	type_method_exception* exp;
	u4 buf4;
	u2 buf2;

	u4* ptr4;
	int i;


	readu4( buf4, cfile );
	readu2( buf2, cfile );

	exp = allocateMemory(
		sizeof( type_method_exception ) + sizeof(u4)*(buf2 - 1 ) );
	exp->number = buf2;
	ptr4 = &exp->exception;

//	printf("exceptions %d %x\n", buf2, exp );

	for( i = 0 ; i < exp->number ; i++ )
	{
		readu2( buf2, cfile );
		ptr4[i] = buf2;
//		exit(0);
	}

	return (exp);
}


void readClassFileAttr( type_class* class, class_file* cfile )
{
	/* not completed */

	return;
}


// ============================================================================
// READ FIELDS
// This function set these
// 	nfield ( number of fields),
// 	fields ( fields table ),
// 	intsize ( field size of instance object )
// ============================================================================

void readFields( type_class* class, class_file* cfile )
{
	type_constant* pool;
	type_field* field;
	u2 buf2;
	u4 buf4;
	u2 count;
	u2 acount;
	int i,j;
	u4 position;
	u4 size;


	readu2( count, cfile );

	if( count <= 0 )
	{
		if( class->super != 0 )
			class->intsize = ((type_class*)class->super)->intsize;
		class->nfield = 0;
		class->fields = 0;
		return;
	}

	pool = class->pool;
	field = (type_field*)allocateMemory( sizeof(type_field) * (count) );

	position = 0;
	size = 0;

	if( class->super != 0 )
		position = ((type_class*)class->super)->intsize;
	
	for( i = 0 ; i < count ; i++ )
	{
		readu2( buf2, cfile );	// flag
		field[i].flag = buf2;
		field[i].cls = class;
		readu2( buf2, cfile );	// name_index
		field[i].name = (pool->data)[buf2];
		readu2( buf2, cfile );	// descriptor
		field[i].descriptor = (pool->data)[buf2];

		field[i].position = 0;
		field[i].size = 0;
		field[i].deprecated = 0;
		field[i].synthetic = 0;
		
		readu2( acount, cfile );

		for( j = 0 ; j < acount ; j++ )
		{
			readu2( buf2, cfile );
			if( (pool->data)[buf2] == string_constant_value )
			{
				readu4( buf4, cfile );
				readu2( buf2, cfile );

				field[i].position = (pool->data)[buf2];
				if( (pool->tag)[buf2] == 0x05 ||
					(pool->tag)[buf2] == 0x06 )
					field[i].size = (pool->data)[buf2+1];
			}
			else if( (pool->data)[buf2] == string_deprecated )
			{
				readu4( buf4, cfile );
				field[i].deprecated = 1;
			}
			else if( (pool->data)[buf2] ==  string_synthetic )
			{
				readu4( buf4, cfile );
				field[i].deprecated = 1;
			}
		}


		if( !(field[i].flag & 0x0008) )
		{
			field[i].position = position;
			
			if( (field[i].descriptor)[0] == 'D' ||
				(field[i].descriptor)[0] == 'J' )
				size = 8;
			else
				size = 4;

			field[i].size = size;
			position += size;
		}
		
//		printf("  [%s] %s %s : %x %x %x\n", class->name,
//			       	field[i].descriptor, field[i].name,
//				field[i].flag, field[i].position, field[i].size );
	}

	class->nfield = count;
	class->fields = field;
	class->intsize = position;
	
	return;
}


void readMethods( type_class* class, class_file* cfile )
{
	type_constant* pool;
	type_method_table* method;
	u2 buf2;
	u4 buf4;
	int nmethod;
	int i,j;
	int acount;
	int parsize, retsize;


	readu2( buf2, cfile );
	nmethod = buf2;

	if( nmethod == 0 )
	{
		class->nmethod = 0;
		class->methods = 0;
		return;
	}

	pool = class->pool;
	method = allocateMemory( sizeof(type_method_table)*nmethod );

	for( i = 0 ; i < nmethod ; i++ )
	{
		readu2( buf2, cfile );	// flag
		method[i].flag = buf2;
		method[i].cls = class;
		readu2( buf2, cfile );	// name_index
		method[i].name = (pool->data)[buf2];
		readu2( buf2, cfile );	// descriptor
		method[i].descriptor = (pool->data)[buf2];

		parseDescriptor( method[i].descriptor,
			&(method[i].parsize), &(method[i].retype) );

		if( (method[i].flag & 0x0008) == 0 )
			method[i].parsize++;

		method[i].code = 0;
		method[i].exception = 0;
		method[i].deprecated = 0;
		method[i].synthetic = 0;

		readu2( acount, cfile );
		
		for( j = 0 ; j < acount ; j++ )
		{
			readu2( buf2, cfile );
			if( (pool->data)[buf2] == string_code )
				method[i].code = readAttributeCode(class, cfile);
			else if( (pool->data)[buf2] == string_exceptions )
				method[i].exception = readAttributeExceptions(class, cfile);
			else if( (pool->data)[buf2] == string_deprecated )
			{
				readu4( buf4, cfile );
				method[i].deprecated = 1;
			}
			else if( (pool->data)[buf2] == string_synthetic )
			{
				readu4( buf4, cfile );
				method[i].synthetic = 1;
			}
		}

		/* debug 
		printf("  [%s] %s %s code=%x\n", class->name,
			       	method[i].name,
			       	method[i].descriptor,
				method[i].code );
		/* debug */

	}

	class->nmethod = nmethod;
	class->methods = method;
	
	return;
}

type_class* readClassData(type_class* class, class_file* cfile)
{
	readVersionInfo( cfile );
	readConstantPool( class, cfile );
	readClassInfo( class, cfile );
	readInterfaces( class, cfile );
	readFields( class, cfile );
	readMethods( class, cfile );
//	readClassFileAttr( class, cfile );
	
	return (class);
}
