#include "type.h"
#include "jtypes.h"
#include "hash.h"


#include "string.h"


hash_t utf8_hash_table;
hash_t string_hash_table;

// =========================================================================
// UTF8 HASH
// =========================================================================

int utf8Hash(const void *ptr)
{
	int hash = 0;
	char* str = ptr;
	int len = strlen( str );


	for( ; len > 0 ; len-- )
		hash = hash * 37 + *str++;

	return (hash);
}


int utf8Comp(const void *ptr1, const void *ptr2)
{
	return (strcmp( ptr1, ptr2 ));
}

// =========================================================================
// STRING HASH
// =========================================================================

int stringHash(const void *ptr)
{
	int hash = 0;
	type_array* array = *(u4*)((type_object*)ptr)->ref;
	jchar* str = array->ref;
	int len = array->length;
	
//	printf(" -> hash {o}%x {a}%x {", ptr, array );
	
	for( ; len > 0 ; len-- )
	{
//		printf("%s", str);
		hash = hash * 37 + *str++;
	}
	
//	printf("} -> ");

	return (hash);
}


int stringComp(const void* *ptr1, const void *ptr2)
{
	type_array* array1 = *(u4*)((type_object*)ptr1)->ref;
	type_array* array2 = *(u4*)((type_object*)ptr2)->ref;
	char* str1 = array1->ref;
	char* str2 = array2->ref;
	int len = ( array1->length >= array2->length )
		? array2->length : array1->length ;
	int result;


	result = strncmp( str1, str2, sizeof(jchar)*len );

	if( array1->length == array2->length && result == 0 )
		return (0);
	else
		return (1);
}


// =========================================================================
// UTF8
// =========================================================================

int initializeStringManager(void)
{
	utf8_hash_table = hashInit( utf8Hash, utf8Comp, 0, 0 );
	string_hash_table = hashInit( stringHash, stringComp, 0, 0 );
	return (1);
}
	
char* utf8New( char* str )
{
	void* found;
	int size;
	
	if( str == 0 ) return (0);

	if( ( found = hashFind( utf8_hash_table, str ) ) == 0 )
	{
		found = malloc( strlen(str)+1 );
		strcpy( found, str );
		found = hashAdd( utf8_hash_table, found );
	}

	return (found);
}

type_object* internString(type_object* str)
{
	type_object* found;

	
	if( str == 0 ) return (0);

	if( ( found = hashFind( string_hash_table, str ) ) == 0 )
		found = hashAdd( string_hash_table, str );
	
	return (found);
}


// =========================================================================
// STRING CONTROL
// =========================================================================

void parseDescriptor( u1* desc, u1* parsize, u1* retype )
{
	int ps, rs;

	

	ps = 0;

//	printf( "\t(parse %s) ", desc);
	
	if( *desc++ != '(' )
		exit(0);

	// count number of parameters

	while( *desc != ')' )
	{
		switch( *desc )
		{
		case 'J':
		case 'D':
			
			ps+=2;
			break;
			
		case '[':

			while( *desc == '[' )
			{
				desc++;
			}

		default:

			if( *desc == 'L' )
			{
				while( *desc != ';' )
					desc++;
			}
			
			ps++;
			break;
		}

		desc++;
	}

	desc++;

	// count size of return

	*retype = *desc;
	*parsize = ps;

	return;
}


unsigned int getUnicodeLength( u1* src )
{
	unsigned int length;
	u1* str;


	str = src;
	length = 0;

	while( *str != '\0' )
	{
		if( (*str&0x80) == 0 )
			str++;
		else if( (*str&0xe0) == 0xc0 )
			str+=2;
		else if( (*str&0xf0) == 0xe0 )
			str+=3;
		else
		{
			printf("unicode error : %x \n", *str );
			exit(0);
		}

		length++;
	}

	return (length);
}

unsigned int getUtf8Length( jchar* uni, unsigned int length )
{
	int result = 0;

	
	while( length-- > 0 )
	{
		if( *uni <= 0x007f )
			result++;
		else if( *uni <= 0x07ff )
			result+=2;
		else if( *uni <= 0xffff )
			result+=3;
		else
			exit(0);
		uni++;
	}
	return (result);
}


char* convertUtf8( jchar* uni, unsigned int length )
{
	unsigned int ulength = getUtf8Length( uni, length );
	u1* u8ch = malloc( sizeof( u1) * (ulength+1) );
	u1* u8start = u8ch;

//	printf( "ulength=%d\n", ulength );
	
	while( length-- > 0 )
	{
//		printf("%c\n",*uni);
		
		if( *uni <= 0x007f )
			*u8ch++ = *uni;
		else if( *uni <= 0x07ff )
		{
			*u8ch++ = ((*uni)&0x07c0) >>6 | 0xc0 ;
			*u8ch++ = ((*uni)&0x003f) | 0x80 ;
		}
		else if( *uni <= 0xffff )
		{
			*u8ch++ = ((*uni)&0xf000) >>12 | 0xd0 ;
			*u8ch++ = ((*uni)&0x0fc0) >>6 | 0x80 ;
			*u8ch++ = ((*uni)&0x003f) | 0x80;
		}
		else
			exit(0);
		uni++;
	}
	*u8ch = '\0';
	return (u8start);
}


void convertUnicode( u1* src, jchar* dst, unsigned int length )
{
	int i;
	unsigned char x, y, z;
	unsigned short tmps;


	for( i = 0 ; i < length ; i++ )
	{
		if( ((*src)&0x80) == 0 )
		{
			dst[i] = (*src) & (0x007f);
			src++;
		}
		else if( ((*src)&0xe0) == 0xc0 )
		{
			x = *src;
			y = *(src+1);
			tmps = ((x&0x1f)<<6) + (y&0x3f);
			dst[i] = tmps;
			src+=2;
		}
		else if( ((*src)&0xf0) == 0xe0 )
		{
			x = *src;
			y = *(src+1);
			z = *(src+2);
			dst[i] = ((x&0xf)<<12 ) + ((y&0x3f)<< 6) + (z&0x3f);
			src+=3;
		}
		else
		{
			printf("convert unicode error\n");
			exit(0);
		}
	}

	return;

}


void restoreJchar( char* dst, jchar* src )
{
	return (0);
}


char* convertNativeFuncName( char* class, char* method, char* sign)
{
	char* tmp;
	char* tmp2;
	char* real;
	char* msig;
	int size;
	int count;
	int i;
	
	
	for( i = strlen(sign) - 1 ; sign[i] != ')' ; i-- );
	
	msig = (char*)malloc( sizeof(char) * i, 0 );
	strncpy(msig,sign+1,i-1);
	msig[i-1] = '\0';

	size = 9 + strlen(class) + strlen(method) + strlen(msig);
	tmp = (char*)malloc( sizeof(char) * size, 0 );
	
	if( msig[0] == '\0' )
		sprintf(tmp,"Java/%s/%s",class,method);
	else
		sprintf(tmp,"Java/%s/%s//%s",class,method,msig);

	for( i = 0, count = 0 ; i < size ; i++ )
	{
		if( tmp[i] == '_' || tmp[i] == ';' || tmp[i] == '[' )
			count++;
	}
	tmp2 = real = (char*)malloc( sizeof(char) * (size + count), 0 );
	
	for( i = 0 ; i < size ; i++ )
	{
		switch( tmp[i] )
		{
		case '_':
			
			*(tmp2++) = '_';
			*(tmp2++) = '1';
			break;

		case ';':

			*(tmp2++) = '_';
			*(tmp2++) = '2';
			break;

		case '[':

			*(tmp2++) = '_';
			*(tmp2++) = '3';
			break;

		case '/':

			*(tmp2++) = '_';
			break;

		default:

			*(tmp2++) = tmp[i];
			break;
		}
	}

	*tmp2 = '\0';

	return (real);	
}


type_array* convertParameters( u1* descriptor, type_loader* loader )
{
	char* current = descriptor;
	type_class* class;
	int count;
	type_array* result;
	char* start;
	char name[100];
	int i = 0;


	if( *current != '(' )
		return (0);

	class = cload( loadUtf8("java/lang/Class"), loader );

	for( count = 0 ; *(++current) != ')' ; count++ )
	{
		if( *current == '[' )
			while( *(++current) == '[' );
		if( *current == 'L' )
			while( *(++current) != ';' );
	}

//	printf("\tparameters counter %d\n", count );

	result = newObjectArray( class, count );

	current = descriptor+1;

	while( *current != ')' )
	{
		switch( *current )
		{
		case '[':

			start = current;
			while( *(++current) == '[' );
			if( *current == 'L' )
				while( *(++current) == ';' );
			strncpy( name, start, (++current) - start );
			name[ current - start ] = '\0';
			break;

		case 'L':

			start = current + 1;
			while( *(++current) != ';' );
			strncpy( name, start, (++current) - start );
			name[ current - start - 1 ] = '\0';
			break;

		default:

			strncpy( name, current, 1 );
			name[1] = '\0';
			break;
		}

		((u4*)result->ref)[i] = cload( loadUtf8(name), loader );
		i++;

//		printf("\t-parameter %d {%s}\n", i, name );
	}
	
	return (result);
}

/*
u1* restroreParameters( type_array* array )
{	
	type_class* class;
	u1 descriptor[256];
	u1* name;

	strcpy( descriptor, "(" );	
	for( i = 0 ; i < array->length ; i++ )
	{
		class = ((u4*)array->ref)[i];
		name = class->name;

		if( name[1] != '\0' )
		{
			strcat( descriptor, "L" );
			strcat( descriptor, name );
			strcat( descriptor, ";" );
		}
		else
		{
			strcat( descriptor, name );
		}
	}
	strcat( descriptor, ")V" );

	return (interUTF8new( descriptor ));
}
*/
