#ifndef __TYPE_H
#define __TYPE_H


#include <stdio.h>
#include "./typendata/macros.h"

typedef unsigned char	u1;
typedef unsigned short	u2;
typedef unsigned int	u4;
typedef unsigned long	u8;

//
#define ACC_PUBLIC		0x0001
#define ACC_PRIVATE		0x0002
#define ACC_PROTECTED		0x0004
#define ACC_STATIC		0x0008
#define ACC_FINAL		0x0010
#define ACC_SUPER		0x0020
#define ACC_VOLATILE		0x0040
#define ACC_TRANSIENT		0x0080
#define ACC_NATIVE		0x0100
#define ACC_INTERFACE		0x0200
#define ACC_ABSTRACT		0x0400


//
#define CONSTANT_UNKNOWN	0x00
#define CONSTANT_ERROR		0x00
#define CONSTANT_UTF8		0x01
#define CONSTANT_INT		0x03
#define CONSTANT_FLOAT		0x04
#define CONSTANT_LONG		0x05
#define CONSTANT_DOUBLE		0x06
#define CONSTANT_CLASS		0x07
#define CONSTANT_STRING		0x08
#define CONSTANT_FIELDREF	0x09
#define CONSTANT_METHODREF	0x0a
#define CONSTANT_INTERFACEMETHODREF	0x0b
#define CONSTANT_NAMEANDTYPE	0x0c

//
#define CONSTANT_UNRESOLVED	0x10
#define CONSTANT_RESOLVED	0x0f

// MACRO
#define IS_UNRESOLVED(p,i)	(((p)->tag)[i] & 0x10)
#define IS_RESOLVED(p,i)	!(((p)->tag)[i] & 0x10)

//
#define DEF_CLASSLIBRARY	"classes.zip"

//
#define PATH_INVALID		0
#define PATH_ZIP		1
#define PATH_DIR		2
#define PATH_SOURCEFILE		3

//
#define CLASSSTATE_UNLOADED	1
#define CLASSSTATE_LOADED	2
#define CLASSSTATE_RESOLVED	3
#define CLASSSTATE_COMPLETED	4



#define STRING_CONSTANT_VALUE	"ConstantValue"
#define STRING_DEPRECATED	"Deprecated"
#define STRING_SYNTHETIC	"Synthetic"
#define STRING_CODE		"Code"
#define STRING_EXCEPTIONS	"Exceptions"

// ============================================================================
// object structure
// ============================================================================

// common object
typedef struct _type_object
{
	struct type_class*	cls;
	u1*			ref;
//	void*			lock;
//	int			count;
} type_object;

// array
typedef struct _type_array
{
	struct type_class*	cls;
	u1*			ref;
	unsigned int		length;
	unsigned int		tsize;
} type_array;

// ============================================================================
// class structure
// ============================================================================

typedef struct _type_class
{
	/* type_object */
	struct type_class*	cls;
	u1*			ref;

	/* type_class */
	struct type_class*	next;
	int			state;
	
	struct type_constant*	pool;

	u1*			name;
	u1*			src_name;
	u2			acc_flag;
	
	struct type_class*	super;
	struct type_loader*	loader;
	int			ninterface;
	u4*			interfaces;

	u4			nfield;
	struct type_field*	fields;
	u4			intsize;
	
	u4			nmethod;
	struct type_method_table* methods;
} type_class;


// ============================================================================
// class loader structure
// ============================================================================

typedef struct _type_loader
{
	type_class	cls;

	/* type_loader */
	u1*			name;
	struct type_class*	list;	// loaded class list in loader
	struct class_pathlist*	path;
	struct type_loader*	super;	
} type_loader;


// constant pool
typedef struct _type_constant
{
	u4			size;
	u4*			data;
	u1*			tag;
} type_constant;

// utf8
typedef struct _utf8
{
	u4			hash;
	u4			size;
	u1*			data;
} utf8;

// field table
typedef struct _type_field
{
	u4			flag;
	struct type_class*	cls;
	u1*			name;
	u1*			descriptor;
	u4			position;
	u4			size;
	u4			deprecated;
	u4			synthetic;
} type_field;

/*
typedef struct _type_static
{
	u4			flag;
	struct type_class*	cls;
	u1*			name;
	u1*			descriptor;
	u8			data;
	u4			deprecated;
	u4			synthetic;
} type_static;
*/

// method table
typedef struct _type_method_table
{
	u4			flag;
	struct type_class*	cls;
	u1*			name;
	u1*			descriptor;
	struct type_method*	code;
	struct type_method_exception*	exception;
	u4			deprecated;
	u4			synthetic;
	u4			parsize;
	u4			retype;
} type_method_table;

// frame templete
typedef struct _type_method
{
	u2			stack;
	u2			local;
	u1*			code;
	u2			exception_length;
	struct type_code_exception*	exception;
	u1*			variable;
	u1*			line;
} type_method;


// ============================================================================
// java frame structure
// ============================================================================

typedef struct _type_frame
{
	struct type_constant*	pool;
	struct type_method_table* method;
	struct type_frame*	prv;
	struct type_object* exception;
	u4* local;
	u4* tos;
	u4* stack;
	u1* base;
	u1* pc;

	/* debug */
	u4 lsize;
	u4 ssize;
} type_frame;


// ============================================================================
// exception structure
// ============================================================================


// exception table
typedef struct _type_method_exception
{
	u2			number;
	u4*			exception;
} type_method_exception;


typedef struct _type_code_exception
{
	u2	from;
	u2	to;
	u2	target;
	u2	type;
} type_code_exception;


// ============================================================================
// thread structure
// ============================================================================

typedef struct _type_native_thread
{
	type_frame* frame;
	type_object* thread;
} type_native_thread;


// ============================================================================
//  structures in class file
// ============================================================================

/* unknowned */
typedef struct _classpath_list
{
	int			type;
	char*			path;
	struct jFILE*		link;
	struct classpath_list*	next;
} classpath_list;

/* unknowned */
typedef struct _class_file
{
	u1*			buf;
	FILE*			fd;
	int			size;
	int			type;
} class_file;


#endif
