
#ifndef DATA_STRUCTURE
#define DATA_STRUCTURE

#include <stdio.h>
#include "./define.h"
#include "../native/jni.h"
//#include "./rules.h"

/////////////////////
/* class structure */
/////////////////////

// + [ Basic Feature ]
// - header
// - list pointer
// - class name
// - source file name
// - access flags
// - super class pointer
// - loader pointer
// - class state
// - gc layout pointer
//-------------------------------
// + [ Constant pool ]
// - constant pool structure
//-------------------------------
// + [ Method Area ]
// - method pointer
// - number of method
// - size of method
//-------------------------------
// + [ Field Area ]
// - field pointer
// - size of non-static fields
// - number of fields
// - number of static fields
// - dispatch table pointer
//-------------------------------
// + [ Interface Area ]
// - pointer of interface structures
// -
// -
// - length of interfaces
// - total length of interfaces
// - all classess implementation of this interface
// - implementation index

/* */
struct Hjava_lang_Object_struct{        
	//struct _dispatchTable*  dtable;
	Hjava_lang_Class*	cls;		// instance of class
        //struct _iLock*          lock;
	void* lock;
	int count;
        /* Data follows on immediately */

//	pthread_mutex_t		mux;
//	pthread_cond_t 		con;

	/* modified 2004/02/17 by numenor */
	runnable_mapping_table* mtable;

	/* modified by numenor */
	field_list*		field;
};

struct Hjava_lang_Class_struct {

	Hjava_lang_Object	base;
	/* header */ 


	utf8 		*name;		//Ŭ ̸ 
	utf8		*src_name;	//  ̸ 
	u2		acc_flags;	//   

	Hjava_lang_Class* super;
	Hjava_lang_ClassLoader*	loader;

	int	state; // unit block is unsigned type;;
	
	constant_pool*	pool;

	methods*	method;
	u2 		nmethods;
	u2		method_sz;
	

	fields*	field;
	u2 fields_cnt;

	int static_field_cnt;
	field_list* static_field;

	/* modified 2004/02/17 by numenor */
	int instance_field_cnt;
	field_list* instance_field;

	/* interface area */
	u4* ifs_table;
	int ifs_num;
	methods* finalizer;
	
	u2 attri_cnt;

	attribute * attributes;
};

//Hjava_lang_Class* new_class();

///////////////////////////////////////////////////////////////
// class loader structure
///////////////////////////////////////////////////////////////
// - base object
// - loaded classes' hash table hendler
// - security protection domain hendler
// - parent class loader pointer
///////////////////////////////////////////////////////////////

struct Hjava_lang_ClassLoader_struct {
	/*
	 * modified by numenor
	 * This object was wanted for REAL JAVA INSTANCE
	 */
	Hjava_lang_Object	base;

	char* name;
	class_list*	list;
	classpath_list* path;
	// protection domain
	Hjava_lang_ClassLoader *super;
};

//////////////////////////////////////////////////////////////
// thread structure
/////////////////////////////////////////////////////////////
#define java_lang_Thread_MAX_PRIORITY 10

/*
struct Hjava_lang_Thread_struct {
	Hjava_lang_Object	base;

//#define java_lang_Thread_MIN_PRIORITY 1
//#define java_lang_Thread_NORM_PRIORITY 5
#define java_lang_Thread_MAX_PRIORITY 10
//	HArrayOfChar* name;
	jint priority;
//	struct Hjava_lang_Thread* threadQ;
	struct Hkaffe_util_Ptr* PrivateInfo;
	jboolean daemon;
//	jboolean interrupting;
//	struct Hjava_lang_Runnable* target;
//	struct Hjava_lang_ThreadGroup* group;
//	struct Hkaffe_util_Ptr* exceptPtr;
//	struct Hjava_lang_Throwable* exceptObj;
//	struct Hkaffe_util_Ptr* jnireferences;
///	struct Hjava_lang_Throwable* stackOverflowError;
//	struct Hjava_lang_Throwable* outOfMemoryError;
//	jboolean started;
//	jboolean dying;
//	struct Hjava_util_HashMap* threadLocals;
//	struct Hjava_lang_Object* suspendResume;
//	struct Hjava_lang_Object* sleeper;
//	struct Hjava_lang_Object* holder;
//	struct Hkaffe_util_Ptr* sem;
//	struct Hjava_lang_Thread* nextlk;
//	struct Hjava_lang_Throwable* death;
//	jint needOnStack;

	pthread_key_t	data_id;
	pthread_key_t	thread_id;

	
	Hjava_lang_Object*	excep;
	runnable*		frame;	// used	
	Hjava_lang_ClassLoader*	loader;	// used
	JNIEnv*                 env;	// used
};
*/

/***************************************************************************
 * Class List
 *
 ***************************************************************************/
#define CLASS_NAME(cl)	((cl)->name->data)

struct class_list_struct {
//	Hjava_lang_ClassLoader	*loader;
	Hjava_lang_Class *cls;
	class_list	*next;
};

//static class_list *class_index[HASH_SIZE] = {0};

/*
class_list * AddClassListEntry(Hjava_lang_Class *cl, Hjava_lang_ClassLoader *loader);
void UpdateClassListEntry(Hjava_lang_Class *cl, Hjava_lang_ClassLoader *loader);
void DeleteClassListEntry(Hjava_lang_Class *cl, Hjava_lang_ClassLoader *loader);
Hjava_lang_Class * FindClassInList (char* class_name, Hjava_lang_ClassLoader *loader);
*/

/***************************************************************************
 * Class path List
 *
 ***************************************************************************/
// path tags.
#define	PATH_INVALID	0
#define	PATH_ZIP	1
#define	PATH_DIR	2
#define	PATH_SOURCEFILE	3

// path list.
struct classpath_list_struct {
	int type;
	char* path;
	jFILE* link; // for jar file
	classpath_list* next;
};


/***************************************************************************
 * Class File Structure
 *
 ***************************************************************************/
struct class_file_struct {
//	u1*	base; // base ptr
	u1*	buf;  // cursor ptr
	FILE *fd;
	int	size; // class file size
	int	type; // class file type : DIR=0, ZIP=1
};

/***************************************************************************
 * Constant Pool Structure
 *
 ***************************************************************************/
struct constant_pool_struct {
	u2	size;
	u1*	tags;
	u8*	data;
} ;

/***************************************************************************
 * UTF8
 *
 ***************************************************************************/
struct utf8_struct {
	u4 	 hash_code;
	u4	 num_of_ref;
	u1	 *data;
};

/***************************************************************************
 * Jar
 *
 ***************************************************************************/

struct _jar {
	struct _jar*	next;
	char*			name;
	long			cur;		// data position
	u2				comp_meth;
	u4				comp_size;
	u4				ucomp_size;
	u4				crc;
};
struct _jFILE {
	class_file*		fp;
	int				cnt;		// entry count
	jar*			head;
	long			cen_entry;	// central dir header offset
	long			loc_entry;	// local file header offset
};

//utf8* utf8_new(char* _str /*=NULL*/, int len);
//void utf8_release(utf8* _utf8_string);
//void utf8_addrefcnt(utf8* _utf8);

//#define	utf8_addrefcnt(_utf8)	++_utf8->num_of_ref

//////////////////////////
//  info  struct 
/////////////////////////

/* fields info */
struct field_info_struct {
        u2	access_flags;
        u2	name_idx;
        u2	type_idx;
	u2	attri_cnt;
};

// method_info_struct 
struct method_info_struct {
        u2	access_flags;
        u2	name_idx;
        u2	type_idx;
	u2	attri_cnt;
};

//////////////////////////
// attribute struct 
/////////////////////////

// used field struct 
struct attributes_constant_value_struct{

	utf8 * attribute_name;
	int attribute_len;
	void * constant_value;    // field descriptor  Ÿ ˷ָ ,  Ÿ ĳϿ ͸  ؾ 

};

// used attribute_code
struct attri_line_number_table_struct
{
	utf8 * name;
	u4 attri_len;
	u2 count;
	struct line_number_info_struct
	{
		u2 start_pc;
		u2 line_number;
	}*line_numbers ;    
};

struct attri_local_variable_table_struct
{
	utf8 * name;
	u4 attri_len;
	u2 count;
	struct local_variable_info_struct
	{
		u2 start_pc;
		u2 length;
		utf8 * name;
		utf8 * type;
		u2 index;
	}*local_variable;
};

//used methode struct 
struct attri_code_struct{
	utf8 * name;
	u4 attri_len;
	u2 max_stack;
	u2 max_locals;
	u4 code_length;
	u1 * code;  // code_length ŭ ޸𸮸 Ҵְ , ͸ Է½Ų. 
	u2 exception_table_len;

	struct exception_table_struct {
		u2 start_pc;
		u2 end_pc;
		u2 handler_pc;
		u2 catch_type;
	}* ex_table;   // exception table ŭ ޸ Ҵ 

	u2 attribute_cnt;	

	struct attr_in_code_struct {
		u1 tag;			 
		attri_local_variable  *    local_variable_table ;
		attri_line_number     *    line_number_table;
	} *attr;

	// line number and local variable
};

struct attri_exception_struct{
	utf8 * name;
	u4 length;
	u2 exception_count;
	u2 *exception_index_table;
};

struct attri_sourcefile_struct{
	utf8* name;
	u4 length;
	utf8 * srcfile_name;
};


struct attri_innerclass_struct{
	utf8* name;
	u4 length;
	u2 num_of_class;

	struct innerclass_table{
		u2 inner_class;
		u2 outer_class;
		utf8 * inner_name;
		u2 access_flag;
	}* class_table;
};


struct attribute_struct{

	u1 tag;			
	attri_sourcefile *sourcefile;
	attri_innerclass *innerclass;
};



///////////////////////////////////////////////////////////////
// field structure
///////////////////////////////////////////////////////////////
// - field name
// - field type
// - access flags
// - field size
// - field attribute info
///////////////////////////////////////////////////////////////

struct fields_struct {
	utf8 * name;
	utf8 * type;
	u2 acc_flag;
    u2 attribute_cnt;

    attri_constant * field_attribute;
};


////////////////////////////////////////////////////////////////
// exception_table structure
////////////////////////////////////////////////////////////////
// - start program counter
// - end program counter
// - exception handler
// - catch_type
////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////
// method structure
///////////////////////////////////////////////////////////////



struct methods_struct {

	u2 acc_flags;
	utf8 * name;
	utf8 * type;
	u2 attribute_cnt;
	Hjava_lang_Class* cls;		// add

	struct attr_in_method_struct 
	{
		u1 tag;					

		attri_code		* code ;
		attri_exception * exception;
	
	}* method_attr;
};

//
// execution runnable instance structure
//
struct _runnable {
	char			*meth_name;
	Hjava_lang_Class	*cls;
	methods			*method;
	u4			*tos;	// top of stack
	u4			*bos;	// base of stack
	constant_pool		*pool;	// constant pool
	u1			*base;	// code base
	u1			*pc;	// program counter
	u4			*local;	// local variable area
	u4			local_size;
	int			code_len;// bytecode length
//	exception_table_struct	*excep;	// exception table
	u2			*excep;
//	struct exception_table_struct {
//		u2 start_pc;
//		u2 end_pc;
//		u2 handler_pc;
//		u2 catch_type;
//	}* excep;
	
	int			excep_len;
	struct _runnable	*prv;	// link

	/* */
	runnable_mapping_table*	mtable;
};

Hjava_lang_ClassLoader* new_loader( Hjava_lang_ClassLoader* super);

/* */
struct runnable_mapping_struct{
	Hjava_lang_Object*			object;
	field_list*				instance;
	field_list*				static_field;
	struct runnable_mapping_struct*		next;
};

//
// object's instance structure
//

struct field_alloc_struct{
	char			*field_name;
	// added by numenor
	char			*member;
	u4			pos;
	u4			size;
};

struct field_list_struct{
	field_alloc     	*field_info;
	u4			*list_ref;
};


#endif
